CREATE OR REPLACE PACKAGE BODY "HRM_SALARY_RULES" is

  g_salessalarycap number:= 4;
  g_logger         varchar2(120);
  g_emailer        varchar2(120);

  

  g_observers Observer_Pattern_Helper.observer_table_t := Observer_Pattern_Helper.observer_table_t();
  
 procedure register_observer
  (  p_observer   in varchar2              
  ,  p_event_type in varchar2 default null -- when null, then observe all event types  
  ) is
  begin
   Observer_Pattern_Helper.register_observer
  ( p_observer   => p_observer
  , p_event_type => p_event_type
  , p_observers  => g_observers
  );
  end register_observer;  
  
  procedure reg_val_fail_observer
  (  p_observer   in varchar2 
  ) is
  begin
    register_observer
    ( p_observer => p_observer
    , p_event_type => C_VALIDATION_FAILURE_EVENT
    );
  end reg_val_fail_observer;

  procedure reg_perf_loss_observer
  (  p_observer   in varchar2 
  ) is
  begin
    register_observer
    ( p_observer => p_observer
    , p_event_type => C_PERFORMANCE_LOSS_EVENT
    );
  end reg_perf_loss_observer;

  procedure reg_sal_decrease_observer
  (  p_observer   in varchar2 
  ) is
  begin
    register_observer
    ( p_observer => p_observer
    , p_event_type => C_SALARY_DECREASE_EVENT
    );
  end reg_sal_decrease_observer;
  
  procedure on_event
  ( p_event_type in varchar2
  , p_event_parameters in Observer_Pattern_Helper.hashmap
  ) is
  begin
    Observer_Pattern_Helper.notify_observers
    ( p_event_type => p_event_type
    , p_event_parameters  => p_event_parameters
    , p_observable => 'HRM_SALARY_RULES'
    , p_observers => g_observers
    );
  end on_event;
  
  procedure set_logger (p_logger in varchar2)
  is
  begin
    g_logger := p_logger;
  end set_logger;
  
  procedure set_emailer (p_emailer in varchar2)
  is
  begin
    g_emailer := p_emailer;
  end set_emailer;
  
  procedure log
  ( p_log_message in varchar2
  , p_log_level   in varchar2 default 'debug'
  , p_log_origin  in varchar2 default 'HRM_SALARY_RULES'
  ) is 
  begin
    if g_logger is not null
    then
      execute immediate 
      'begin
         '||g_logger
         ||' ( p_log_level => :1
             , p_log_origin => :2
             , p_log_message => :3
             );
          end;'
       using in p_log_level
       ,     in p_log_origin
       ,     in p_log_message
       ;
    end if;
  end log; 

procedure send_email
 ( p_sender     in varchar2
 , p_recipient  in varchar2
 , p_header     in varchar2
 , p_message    in varchar2
 , p_importance in varchar2
 ) is
 begin
    if g_emailer is not null
    then
      execute immediate 
      'begin
         '||g_emailer
         ||' ( p_sender     => :1
             , p_recipient  => :2
             , p_header     => :3
             , p_message    => :4
             , p_importance => :5
             );
          end;'
       using in p_sender
       ,     in p_recipient
       ,     in p_header
       ,     in p_message
       ,     in p_importance
       ;
    end if;
 end send_email;
 
  procedure set_salessalarycap( p_salessalarycap in number)
  is
  begin
    g_salessalarycap:= p_salessalarycap;
  end set_salessalarycap;

  function verify_salary_raise( p_job in varchar2, p_old_salary in number, p_new_salary in number)
  return boolean
  is
    l_start_time number:= dbms_utility.get_time; -- returns a timestamp in 100ths of a second
    l_end_time   number;
    l_result     boolean:= true;
  begin 
    log('verify_salary_raise for job='||p_job||' and old salary '||p_old_salary||' changing into new salary '||p_new_salary, p_log_level=> 'info');
    if p_old_salary > p_new_salary
    then
      on_event
      ( C_SALARY_DECREASE_EVENT
      , Observer_Pattern_Helper.key_value_string_to_hashmap
        ( 'old_salary='||p_old_salary||';'
        ||'new_salary='||p_new_salary||';'
        ||'user='||USER||';'
        )
      );
    end if;
    dbms_lock.sleep(0.5); -- deliberate performance degradation
    -- among a lot of other tests...
    if instr(lower(p_job),'sales') > 0
    then
      log( '- salarycap = '||g_salessalarycap);
      if (p_new_salary - p_old_salary)/p_old_salary > g_salessalarycap/100
      then
        log('- FALSE: this change in salary for this job is not in accordance with the rules', p_log_level=> 'info');
       
        send_email
        ( p_sender => 'jellema@amis.nl'
        , p_recipient => 'jellema@amis.nl'
        , p_header => 'Attempt to violate Salary Rule'
        , p_message => 'Someone tried to update a salary of a '||p_job
                     ||' from '||p_old_salary||' to '||p_new_salary
                     ||'  which is illegal!'
        , p_importance => 'HIGH'
        );
        
        l_result:= false;
      end if;
      log('- TRUE: this change in salary for this job is in accordance with the rules');
    end if;
    l_end_time:= dbms_utility.get_time;
    -- test for total validation time; if more than 400 ms, fire event
    if (l_end_time - l_start_time) > 40 -- 40 * one 100th of a second
    then
      on_event
      ( C_PERFORMANCE_LOSS_EVENT
      , Observer_Pattern_Helper.key_value_string_to_hashmap
        ( 'validation_time='||10*(l_end_time - l_start_time)||' ms;'
        )
      );    
    end if;
    if not l_result
    then
      on_event
      ( C_VALIDATION_FAILURE_EVENT
      , Observer_Pattern_Helper.key_value_string_to_hashmap
        ( 'old_salary='||p_old_salary||';'
        ||'new_salary='||p_new_salary||';'
        ||'job='||p_job||';'
        ||'salessalarycap='||g_salessalarycap||';'
        ||'user='||USER||';'
        )
      );
    end if;

    return l_result;
  end verify_salary_raise;

  -- @AOP(advice=LOG)
  function verify_job_change( p_old_job in varchar2, p_new_job in varchar2)
  return boolean
  is
  begin

    -- AOP-ADVICE:LOG  ; Added by AOP_PROCESSOR on 16-03-2006 07:13:45
    declare
      l_parameters CENTRAL_LOGGER.parameter_table;
    begin
      l_parameters('p_old_job'):= p_old_job;
      l_parameters('p_new_job'):= p_new_job;
      CENTRAL_LOGGER.log_program_unit_execution
      ( p_program_unit => 'HRM_SALARY_RULES.verify_job_change'
      , p_parameters => l_parameters
      );
    end;
    -- END OF AOP-ADVICE

                                                                                                                                                                                                                          if p_old_job in ('SALESMAN','CLERK') and p_new_job in ('MANAGER','PRESIDENT')
    then
      return false;
    end if;
    return true;    
  end verify_job_change;
  
end HRM_SALARY_RULES;
